# -*- coding: utf-8 -*-
"""
Custom exceptions for seedir.

"""

# NOTE: SeedirError removed for 0.4.0

# class SeedirError(Exception):
#     """Class for representing errors from module `seedir.realdir`"""

class FakedirError(Exception):
    """Class for handling errors from module `seedir.fakedir`"""



# -*- coding: utf-8 -*-
"""
The primary algorithm for determining the folder structure returned by
`seedir.realdir.seedir()` and `seedir.fakedir.FakeDir.seedir()`.

"""

from abc import ABC, abstractmethod
import copy
import math
import os

import natsort

import seedir.printing as printing

def listdir_fullpath(path):
    '''Like `os.listdir()`, but returns absolute paths.'''
    return [os.path.join(path, f) for f in os.listdir(path)]

class FolderStructureArgs:

    def __init__(self, extend='│ ', space='  ', split='├─', final='└─',
                 filestart='', folderstart='', fileend='', folderend='/',
                 depthlimit=None, itemlimit=None, beyond=None, first=None,
                 sort=True, sort_reverse=False, sort_key=None,
                 include_folders=None, exclude_folders=None,
                 include_files=None, exclude_files=None,
                 regex=False, mask=None, formatter=None,
                 sticky_formatter=False, acceptable_listdir_errors=None,
                 denied_string=' [ACCESS DENIED]'):
        self.extend = extend
        self.space = space
        self.split = split
        self.final = final
        self.filestart = filestart
        self.folderstart = folderstart
        self.fileend = fileend
        self.folderend = folderend
        self.depthlimit = depthlimit
        self.itemlimit = itemlimit
        self.beyond = beyond
        self.first = first
        self.sort = sort
        self.sort_reverse = sort_reverse
        self.sort_key = sort_key
        self.include_folders = include_folders
        self.exclude_folders = exclude_folders
        self.include_files = include_files
        self.exclude_files = exclude_files
        self.regex = regex
        self.mask = mask
        self.formatter = formatter
        self.sticky_formatter = sticky_formatter
        self.acceptable_listdir_errors = acceptable_listdir_errors
        self.denied_string = denied_string

    def copy(self):
        return copy.deepcopy(self)

    def update_with_formatter(self, formatter, item):
        newstyle = formatter(item)
        if newstyle is None:
            return
        for k, v in newstyle.items():
            setattr(self, k, v)

class FolderStructure(ABC):
    '''General class for determining folder strctures.  Implements
    the seedir folder-tree generating algorithm over arbitrary objects.

    **This is an abstract class that cannot be instantiated directly.**  It must
    be subclassed, and the user must imeplement three abstract methods:

        - `getname(self, item)`: Return the string name of a folder or file object.
        - `isdir(self, item)`: Return a boolean indicating whether an object is
        a folder (i.e., can be further travesed) or not.
        - `listdir(self, item)`: For a folder-like object, return its children,
        such that these same three methods can be called on each child again
        (recursively).

    Details on each of these functions are provided below.  In each case,
    `item` is a folder-like or file-like object that can be listed in the
    diagram.

    Seedir provides implementations for string paths (`RealDirStructure`),
    pathlib objects (`PathibStucture`), and "fakedir" objects (`FakeDirStucture`).

    *Note: Prior to v0.5.0, you could initialize a FolderStructure by passing
    these three functions as arguments to the constructor.  This is now
    deprecated and will throw an error.*
    '''

    # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
    # ABSTRACT METHODS WHICH REQUIRE IMPLEMENTATION
    # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

    @abstractmethod
    def getname(self, item):
        '''
        Generate a string name for an item to put in the tree diagram.

        Parameters
        ----------
        item : user-specified
            Folder-like or file-like object.

        Returns
        -------
        name : str
            Name to show in the diagram.

        '''
        ...

    @abstractmethod
    def isdir(self, item):
        '''
        Returns True if an object is folder-like.

        This function determines whether to treat objects as a folder or a
        file.  Some key differences of note are:

        - Folders **will** be passed to `listdir()` to retrieve their
        children.  Files are instead not passed to this function.
        - Some arguments are unique to folders or files, e.g.
        `exclude_folders` and `exclude_files`.
        - Folders and files are (by default) represented differently
        in diagrams, with folders usually ending with a slash.

        Parameters
        ----------
        item : user-specified
            Folder-like or file-like object.

        Returns
        -------
        result : bool
            True if item is folder; False if not.

        '''
        ...

    @abstractmethod
    def listdir(self, item):
        '''
        Return the children of a folder-type object.

        Children should always be more objects which can be passed to
        `getname()`, `isdir()`, or (in the case of folder-like objects)
        `listdir()`.

        This function is not called on objects which are not folders,
        as determined by `isdir()`.

        Parameters
        ----------
        item : user-specified
            Folder-like or file-like object.

        Returns
        -------
        children : list
            List of child objects for further traversal.  Can be empty
            for a folder with no children.

        '''
        ...

    # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

    def __call__(self, folder, style='lines', printout=True, indent=2, uniform=None,
                 anystart=None, anyend=None, depthlimit=None, itemlimit=None,
                 beyond=None, first=None, sort=False, sort_reverse=False,
                 sort_key=None, include_folders=None, exclude_folders=None,
                 include_files=None, exclude_files=None, regex=False, mask=None,
                 formatter=None, sticky_formatter=False,
                 acceptable_listdir_errors=None, denied_string='', **kwargs):
        '''Call this on a folder object to generate the seedir output
        for that object.'''

        accept_kwargs = ['extend', 'split', 'space', 'final',
                         'folderstart', 'filestart', 'folderend', 'fileend']

        if any(i not in accept_kwargs for i in kwargs.keys()):
            bad_kwargs = [i for i in kwargs.keys() if i not in accept_kwargs]
            raise ValueError(f'kwargs must be any of {accept_kwargs}; '
                             f'unrecognized kwargs: {bad_kwargs}')

        styleargs = printing.get_styleargs(style)
        printing.format_indent(styleargs, indent=indent)

        if uniform is not None:
            for arg in ['extend', 'split', 'final', 'space']:
                styleargs[arg] = uniform

        if anystart is not None:
            styleargs['folderstart'] = anystart
            styleargs['filestart'] = anystart

        if anyend is not None:
            styleargs['folderend'] = anyend
            styleargs['fileend'] = anyend

        for k in kwargs:
            if k in styleargs:
                styleargs[k] = kwargs[k]

        if sort_key is not None or sort_reverse == True:
            sort = True

        if acceptable_listdir_errors is None:
            acceptable_listdir_errors = tuple()

        args = FolderStructureArgs(depthlimit=depthlimit,
                                   itemlimit=itemlimit,
                                   beyond=beyond,
                                   first=first,
                                   sort=sort,
                                   sort_reverse=sort_reverse,
                                   sort_key=sort_key,
                                   include_folders=include_folders,
                                   exclude_folders=exclude_folders,
                                   include_files=include_files,
                                   exclude_files=exclude_files,
                                   regex=regex,
                                   mask=mask,
                                   formatter=formatter,
                                   sticky_formatter=sticky_formatter,
                                   acceptable_listdir_errors=acceptable_listdir_errors,
                                   denied_string=denied_string,
                                   **styleargs)


        s = self._folder_structure_recurse(ITEM=folder,
                                           FSARGS=args).strip()

        if printout:
            print(s)
        else:
            return s

    def apply_itemlimit(self, items, itemlimit):
        '''
        Limit the number of children in a folder.

        Parameters
        ----------
        items : list
            Child items.
        itemlimit : int, tuple, or None
            Limit.  If int, limits the number of items regardless of if they
            are folders or file.  If a 2-tuple, the first number is treated
            as a limit on the number of folders and the second number
            is treated as a limit on the number of files.  If None,
            no limit is applied.

        Returns
        -------
        finalitems : list
            Kept times.
        rem : list
            Omitted items.

        '''

        # all items included when None
        if itemlimit is None:
            itemlimit = len(items)

        # if int, take the first itemlimit items
        if isinstance(itemlimit, int):
            itemlimit = min(itemlimit, len(items))
            finalitems = items[:itemlimit]
            rem = items[itemlimit:]

        # if tuple, interpret as limits for folders and files
        else:
            finalitems, rem = self._apply_tuple_itemlimit(items, itemlimit)

        return finalitems, rem

    def _apply_tuple_itemlimit(self, items, itemlimit):
        '''Apply the itemlimit when it is a 2-tuple of
        (folderlimit, filelimit).'''
        folderlimit, filelimit = itemlimit

        folderlimit = math.inf if folderlimit is None else folderlimit
        filelimit = math.inf if filelimit is None else filelimit

        finalitems = []
        rem = []
        foldercount = 0
        filecount = 0

        for item in items:
            isdir = self.isdir(item)

            if isdir and (foldercount < folderlimit):
                finalitems.append(item)
                foldercount += 1
            elif isdir:
                rem.append(item)
            elif not isdir and (filecount < filelimit):
                finalitems.append(item)
                filecount += 1
            else:
                rem.append(item)

        return finalitems, rem

    def beyond_depth_str(self, items, beyond):
        '''
        Generates the text when using the 'beyond'
        parameter and ther are more items than the itemlimit or contents
        beyond the depthlimit.

        Parameters
        ----------
        beyond : str
            Style of beyond string to generate.  Options are:
                - 'ellipsis' ('...')
                - 'content' or 'contents' (the number of files and folders beyond)
                - a string starting with '_' (everything after the leading
                  underscore will be returned)
        items : collection of file paths, optional
            Path items of the items beyond the limit, used when the 'beyond'
            argeument is 'content' or 'contents'. The default is None.


        Returns
        -------
        str
            String indicating what lies beyond

        '''
        if beyond.lower() == 'ellipsis':
            return '...'
        elif beyond.lower() in ['contents','content']:
            folders = self.count_folders(items)
            files = self.count_files(items)
            return '{} folder(s), {} file(s)'.format(folders, files)
        elif beyond and beyond[0] == '_':
            return beyond[1:]
        else:
            s1 = '"beyond" must be "ellipsis", "content", or '
            s2 = 'a string starting with "_"'
            raise ValueError(s1 + s2)

    def count_files(self, items):
        '''
        Count the number of files in a collection of paths.

        Parameters
        ----------
        paths : list-like
            Collection of paths.

        Returns
        -------
        files : int
            Count of files.

        '''
        files = sum([not self.isdir(i) for i in items])
        return files

    def count_folders(self, items):
        '''
        Count the number of folders in a collection of paths.

        Parameters
        ----------
        paths : list-like
            Collection of paths.

        Returns
        -------
        files : int
            Count of folders.

        '''
        folders = sum([self.isdir(i) for i in items])
        return folders

    def filter_items(self, listdir, include_folders=None,
                      exclude_folders=None, include_files=None,
                      exclude_files=None, regex=False, mask=None):

        '''
        Filter for folder and file names in folder structures.  Removes or includes
        items matching filtering strings.

        Note the following priority of arguments:
            1. Mask (totally overwrites include/exclude)
            2. Include (saves items from being excluded)
            3. Exclude

        Parameters
        ----------
        listdir : list-like
            Collection of file/folder items.
        include_folders : str or list-like, optional
            Folder names to include. The default is None.
        exclude_folders : str or list-like, optional
            Folder names to exclude. The default is None.
        include_files : str or list-like, optional
            File names to include. The default is None.
        exclude_files : str or list-like, optional
            File names to exclude. The default is None.
        regex : bool, optional
            Interpret strings as regular expressions. The default is False.
        mask : function, optional
            Function for filtering items.  Absolute paths of each individual item
            are passed to the mask function.  If True is returned, the
            item is kept.  The default is None.

        Returns
        -------
        keep : list
            Filtered input.

        '''

        def _should_convert(x):
            return isinstance(x, str) or x is None

        filtered = []
        for item in listdir:

            name = self.getname(item)

            if self.isdir(item):
                inc = [include_folders] if _should_convert(include_folders) else include_folders
                exc = [exclude_folders] if _should_convert(exclude_folders) else exclude_folders
            else:
                inc = [include_files] if _should_convert(include_files) else include_files
                exc = [exclude_files] if _should_convert(exclude_files) else exclude_files

            # 1. check mask - which trumps include/exclude arguments
            if mask is not None:
                if mask(item):
                    filtered.append(item)
                continue

            # set default keep behavior
            # items are exluded if inclusion is passed
            keep = all([i is None for i in inc])

            # 2. apply exclusion
            for pat in exc:
                if pat is not None:
                    match = printing.is_match(pattern=pat, string=name, regex=regex)
                    if match:
                        keep = False

            # 3. apply inclusion (trumps exclusion)
            for pat in inc:
                if pat is not None:
                    match = printing.is_match(pattern=pat, string=name, regex=regex)
                    if match:
                        keep = True

            if keep:
                filtered.append(item)

        return filtered

    def _folder_structure_recurse(self, ITEM, FSARGS, DEPTH=0,
                                  INDEX=0, INCOMPLETE=None,
                                  IS_LASTITEM=False,
                                  IS_RAWSTRING=False):

        # initialization
        OUTPUT = ''

        if INCOMPLETE is None:
            INCOMPLETE = []

        # set some variables
        is_rootitem = DEPTH == 0
        is_lastitem = IS_LASTITEM
        is_rawstring = IS_RAWSTRING

        # APPLY FORMATTER
        # # # # # # # # # # # # # #
        args = FSARGS.copy()
        next_args = FSARGS

        if not is_rawstring and args.formatter is not None:
            args.update_with_formatter(args.formatter, ITEM)

        if args.sticky_formatter:
            next_args = args

        # GET CHILDREN
        # # # # # # # # # # # # # #

        error_listing = False

        if not is_rawstring and self.isdir(ITEM):
            try:
                listdir = self.listdir(ITEM)
            except args.acceptable_listdir_errors:
                error_listing = True
                listdir = None
        else:
            listdir = None

        # ADD CURRENT ITEM TO OUTPUT
        # # # # # # # # # # # # # #

        # create header
        base_header = self.get_base_header(INCOMPLETE,
                                           args.extend,
                                           args.space)

        # handle ultimate token in header
        if is_rootitem:
            branch = ''
        elif is_lastitem:
            branch = args.final
        else:
            branch = args.split

        header = base_header + branch

        # start / end tokens
        if is_rawstring:
            fkey = 'file'
        else:
            fkey = 'folder' if self.isdir(ITEM) else 'file'

        start = f'{fkey}start'
        end = f'{fkey}end'

        # add current item to string
        name = self.getname(ITEM) if not is_rawstring else ITEM
        error_tag = args.denied_string if error_listing else ''

        OUTPUT += (header +
                   getattr(args, start) +
                   name +
                   getattr(args, end) +
                   error_tag +
                   '\n')

        if is_lastitem and INCOMPLETE:
            INCOMPLETE.remove(DEPTH-1)

        # EXIT IF NOT FOLDER
        # # # # # # # # # # # # # #

        if listdir is None:
            return OUTPUT

        # FILTER/SORT CHILDREN
        # # # # # # # # # # # # #

        current_itemlimit = args.itemlimit

        # handle when depthlimit is reached
        if isinstance(args.depthlimit, int) and DEPTH >= args.depthlimit:
            if args.beyond is None:
                return OUTPUT
            else:
                current_itemlimit = 0

        # sort and filter the contents of listdir
        sortargs = {
            'first': args.first,
            'sort_reverse': args.sort_reverse,
            'sort_key': args.sort_key}

        filterargs = {
            'include_folders': args.include_folders,
            'exclude_folders': args.exclude_folders,
            'include_files' : args.include_files,
            'exclude_files': args.exclude_files,
            'mask': args.mask,
            }

        if args.sort or args.first is not None:
            listdir = self.sort_dir(listdir, **sortargs)

        if any(arg is not None for arg in filterargs.values()):
            listdir = self.filter_items(listdir, **filterargs, regex=args.regex)

        # apply itemlimit
        finalitems, rem = self.apply_itemlimit(listdir, current_itemlimit)

        # append beyond string if being used
        beyond_added = False
        if args.beyond is not None:
            if rem or (DEPTH == args.depthlimit):
                finalitems += [self.beyond_depth_str(rem, args.beyond)]
                beyond_added = True

        # RECURSE
        # # # # # # # # # # # # # #

        if finalitems:
            INCOMPLETE.append(DEPTH)

        total = len(finalitems)

        for i, x in enumerate(finalitems):
            last = i == (total - 1)
            IS_RAWSTRING = (beyond_added and last)
            OUTPUT += self._folder_structure_recurse(x, DEPTH=DEPTH+1,
                                                     INCOMPLETE=INCOMPLETE,
                                                     FSARGS=next_args,
                                                     INDEX=i,
                                                     IS_LASTITEM=last,
                                                     IS_RAWSTRING=IS_RAWSTRING)


        # RETURN
        # # # # # # # # # # # # # #
        return OUTPUT

    def get_base_header(self, incomplete, extend, space):
        '''
        For folder structures, generate the combination of extend and space
        tokens to prepend to file names when generating folder diagrams.


        The string generated here will still be missing the branch token
        (split or final) as well as any folderstart or filestart tokens.
        They are added within seedir.seedir().

        For any item included in a folder diagram, the combination of
        extend and space tokens is based on the depth of the item as well as the
        parent folders to that item which are not completed.  This information
        is symbolized with the `incomplete` argument.  The following illustrates
        the incomplete arguments passed to this function for an example folder
        tree:

        ```
        doc/
        ├─_static/                  [0]
        │ ├─embedded/               [0, 1]
        │ │ ├─deep_file             [0, 1, 2]
        │ │ └─very/                 [0, 1, 2]
        │ │   └─deep/               [0, 1, 3]
        │ │     └─folder/           [0, 1, 4]
        │ │       └─very_deep_file  [0, 1, 5]
        │ └─less_deep_file          [0, 1]
        └─index.rst                 [0]
        ```

        Parameters
        ----------
        incomplete : list-like
            List of integers denoting the depth of incomplete folders at the time
            of constructing the line for a given item.  Zero represents being
            inside the main folder, with increasing integers meaing increasing
            depth.
        extend : str
            Characters symbolizing the extension of a folder.
        space : str
            Characters providing the gap between items and earlier parents.

        Returns
        -------
        str
            Base header string.

        '''
        if not incomplete:
            return ''

        base_header = []
        max_i = max(incomplete)
        for p in range(max_i):
            if p in incomplete:
                base_header.append(extend)
            else:
                base_header.append(space)
        return "".join(base_header)

    def sort_dir(self, items, first=None, sort_reverse=False, sort_key=None):
        '''
        Sorting function used to sort contents when producing folder diagrams.

        Parameters
        ----------
        items : list-like
            Collection of folder contents.
        first : 'files' or 'folders', optional
            Sort either files or folders first. The default is None.
        sort_reverse : bool, optional
            Reverse the sort applied. The default is False.
        sort_key : function, optional
            Function to apply to sort the objs by their basename.  The function
            should take a single argument, of the type expected by
            this FolderStucture.

        Returns
        -------
        list
            Sorted input as a list.

        '''
        if sort_key is None:
            key = lambda x : self.getname(x)
        else:
            key = lambda x: sort_key(self.getname(x))

        if first in ['folders', 'files']:
            folders = [p for p in items if self.isdir(p)]
            files = [p for p in items if not self.isdir(p)]
            folders = natsort.natsorted(folders, reverse=sort_reverse, key=key)
            files = natsort.natsorted(files, reverse=sort_reverse, key=key)
            output = folders + files if first == 'folders' else files + folders
        elif first is None:
            output = list(natsort.natsorted(items, reverse=sort_reverse, key=key))
        else:
            raise ValueError("`first` must be 'folders', 'files', or None.")

        return output

class RealDirStructure(FolderStructure):
    """Make folder structures from string paths."""

    def __init__(self):
        super().__init__()
        self.slashes = os.sep + '/' + '//'

    def getname(self, item):
        return os.path.basename(item.rstrip(self.slashes))

    def isdir(self, item):
        return os.path.isdir(item)

    def listdir(self, item):
        return listdir_fullpath(item)

class PathlibStructure(FolderStructure):
    """Make folder structures from pathlib objects."""

    def getname(self, item):
        return item.name

    def isdir(self, item):
        return item.is_dir()

    def listdir(self, item):
        return list(item.iterdir())

class FakeDirStructure(FolderStructure):
    """Make `seedir.fakedir.FakeDir` folder structures."""

    def getname(self, item):
        return item.name

    def isdir(self, item):
        return item.isdir()

    def listdir(self, item):
        return item.listdir()



# -*- coding: utf-8 -*-
"""
[seedir](https://github.com/earnestt1234/seedir) is a Python package for
creating, editing, and reading folder tree diagrams.

The general logic of seedir is based on representing directories in 3
different forms: real directories on a computer, text diagrams of directories,
and "fake directories" (manipulable Python folder objects). seedir provides
tools for going in between these formats.

.. include:: ../docs/gettingstarted.md

## Cheatsheet

![](https://raw.githubusercontent.com/earnestt1234/seedir/master/img/seedir_diagram.png)

"""

#imports for package namespace
from .realdir import seedir

from .fakedir import (fakedir,
                      fakedir_fromstring,
                      populate,
                      randomdir,)

from .fakedir import FakeDir
from .fakedir import FakeFile

from .printing import (get_styleargs,
                       STYLE_DICT,)



# -*- coding: utf-8 -*-
"""
Unit tests for seedir.

Previously written for unittests, and updated for pytest.

Test methods MUST start with "test"
"""

import os

import pytest

import seedir as sd
from seedir.errors import FakedirError
from seedir.folderstructure import FakeDirStructure as FDS

# ---- Test seedir strings

example = """mypkg/
    __init__.py
    app.py
    view.py
    test/
        __init__.py
        test_app.py
        test_view.py"""

no_init = """mypkg/
    app.py
    view.py
    test/
        test_app.py
        test_view.py"""

example_with_comments="""mypkg/
    __init__.py #some comments
    app.py #####        more comments
    view.py #how about## this one
    test/
        __init__.py
        test_app.py
        test_view.py"""

# sd.randomdir(seed=456)
large_example = """MyFakeDir/
├─Vogel.txt
├─monkish.txt
├─jowly.txt
├─scrooge/
│ ├─light.txt
│ ├─reliquary.txt
│ ├─sandal/
│ ├─paycheck/
│ │ ├─electrophoresis.txt
│ │ └─Pyongyang/
│ └─patrimonial/
├─Uganda/
└─pedantic/
  └─cataclysmic.txt"""

large_example_access_denied = """MyFakeDir/
├─Vogel.txt
├─monkish.txt
├─jowly.txt
├─scrooge/ [ACCESS DENIED]
├─Uganda/ [ACCESS DENIED]
└─pedantic/ [ACCESS DENIED]"""

large_example_bummer = """MyFakeDir/
├─Vogel.txt
├─monkish.txt
├─jowly.txt
├─scrooge/<- BUMMER!
├─Uganda/<- BUMMER!
└─pedantic/<- BUMMER!"""

limit0_nobeyond = 'MyFakeDir/'

limit0_beyond_content = """MyFakeDir/
└─3 folder(s), 3 file(s)"""

depthlimit1 = '''MyFakeDir/
├─Vogel.txt
├─monkish.txt
├─jowly.txt
├─scrooge/
├─Uganda/
└─pedantic/'''

depthlimit1_beyond_content = '''MyFakeDir/
├─Vogel.txt
├─monkish.txt
├─jowly.txt
├─scrooge/
│ └─3 folder(s), 2 file(s)
├─Uganda/
│ └─0 folder(s), 0 file(s)
└─pedantic/
  └─0 folder(s), 1 file(s)'''

depthlimit1_beyond_content_exclude = '''MyFakeDir/
├─scrooge/
│ └─3 folder(s), 0 file(s)
├─Uganda/
│ └─0 folder(s), 0 file(s)
└─pedantic/
  └─0 folder(s), 0 file(s)'''

complex_sort = '''MyFakeDir/
├─monkish.txt
├─Vogel.txt
├─jowly.txt
├─pedantic/
│ └─cataclysmic.txt
├─scrooge/
│ ├─reliquary.txt
│ ├─light.txt
│ ├─patrimonial/
│ ├─paycheck/
│ │ ├─electrophoresis.txt
│ │ └─Pyongyang/
│ └─sandal/
└─Uganda/'''

complex_inclusion = '''MyFakeDir/
├─Vogel.txt
├─monkish.txt
├─jowly.txt
├─scrooge/
│ ├─light.txt
│ └─sandal/
└─pedantic/'''

fmt_expand_single = '''MyFakeDir/
├─Vogel.txt
├─monkish.txt
├─jowly.txt
├─scrooge/
│ ├─light.txt
│ ├─reliquary.txt
│ ├─sandal/
│ ├─paycheck/
│ │ ├─electrophoresis.txt
│ │ └─Pyongyang/
│ └─patrimonial/
├─Uganda/
└─pedantic/'''

fmt_expand_single_partial = '''MyFakeDir/
├─Vogel.txt
├─monkish.txt
├─jowly.txt
├─scrooge/
│ ├─light.txt
│ ├─reliquary.txt
│ ├─sandal/
│ ├─paycheck/
│ └─patrimonial/
├─Uganda/
└─pedantic/'''

fmt_notbeyond = """MyFakeDir/
->Vogel.txt
└─3 folder(s), 2 file(s)"""

fmt_with_mask = '''MyFakeDir/
├─scrooge/
│ ├─light.txt
│ └─reliquary.txt
├─Uganda/
└─pedantic/
  └─cataclysmic.txt'''

# variables
styles = ['lines', 'dash', 'spaces', 'arrow', 'plus', 'emoji']
try:
    import emoji
except ImportError:
    styles.remove('emoji')

# realdir for testing on
testdir = os.path.dirname(os.path.abspath(__file__))

# custom FakeDirStructure for testing errors
class ErrorRaisingFDS(FDS):

    def listdir(self, item):
        if self.isdir(item) and item.depth != 0:
            raise FakedirError('Oops!!!')
        else:
            return item.listdir()

# ---- Test cases

class PrintSomeDirs:

    print('\n--------------------'
          '\n\nTesting seedir.seedir() against {}:\n\n'
          '--------------------'
          '\n'.format(testdir))
    def test_a_print_userprofile(self):
        print('Basic seedir (depthlimit=2, itemlimit=10):\n')
        sd.seedir(testdir, depthlimit=2, itemlimit=10)

    def test_b_styles(self):
        print('\nDifferent Styles (depthlimit=1, itemlimit=5):')
        for style in sd.STYLE_DICT.keys():
            print('\n{}:\n'.format(style))
            sd.seedir(testdir, style=style, depthlimit=1, itemlimit=5)

    def test_c_custom_styles(self):
        print('\nCustom Styles (depthlimit=1, itemlimit=5):')
        sd.seedir(testdir, depthlimit=1, itemlimit=5, space='>>',
                  split='>>', extend='II', final='->',
                  folderstart='Folder: ', filestart='File: ')

    def test_d_indent(self):
        print('\nDifferent Indents (depthlimit=1, itemlimit=5):')
        for i in list(range(3)) + [8]:
            print('\nindent={}:\n'.format(str(i)))
            sd.seedir(testdir, depthlimit=1, itemlimit=5, indent=i)

    def test_e_beyond(self):
        print('\nItems Beyond Limit (depthlimit=1, itemlimit=1, beyond="content")')
        sd.seedir(testdir, itemlimit=1, beyond='content')

    def test_improper_kwargs(self):
        with pytest.raises(ValueError):
            sd.seedir(testdir, spacing=False)

class TestSeedirStringFormatting:
    def test_get_base_header_0(self):
        a = '| '
        b = '  '
        assert FDS().get_base_header([0], a, b) == ''

    def test_get_base_header_013(self):
        a = '| '
        b = '  '
        assert FDS().get_base_header([0, 1, 3], a, b) == '| |   '

    def test_get_base_header_empty(self):
        a = '| '
        b = '  '
        assert FDS().get_base_header([], a, b) == ''

    def test_STYLE_DICT_members(self):
        keys = set(sd.STYLE_DICT.keys())
        assert keys == set(styles)

    @pytest.mark.parametrize('style', styles)
    def test_get_style_args_all_accessible(self, style):
        d = sd.get_styleargs(style)
        assert isinstance(d, dict)

    def test_access_missing_style(self):
        with pytest.raises(ValueError):
            _ = sd.get_styleargs('missing_style')

    def test_get_style_args_deepcopy(self):
        x = sd.STYLE_DICT['lines']
        y = sd.get_styleargs('lines')
        assert x is not y

    def test_format_indent_4(self):
        a = sd.get_styleargs('lines')
        sd.printing.format_indent(a, indent=4)
        chars = ['extend', 'space', 'split', 'final']
        assert all(len(a[c])==4 for c in chars)

    def test_format_indent_1(self):
        a = sd.get_styleargs('lines')
        sd.printing.format_indent(a, indent=1)
        chars = ['extend', 'space', 'split', 'final']
        assert all(len(a[c])==1 for c in chars)

    def test_words_list_start(self):
        assert sd.printing.words[0] == 'a'

    def test_words_list_length(self):
        assert len(sd.printing.words) == 25487

class TestFakeDirReading:
    def test_read_string(self):
        x = sd.fakedir_fromstring(example)
        assert isinstance(x, sd.FakeDir)

    def test_parse_comments_on(self):
        x = sd.fakedir_fromstring(example)
        y = sd.fakedir_fromstring(example_with_comments)
        assert x.get_child_names() == y.get_child_names()

    def test_parse_comments_off(self):
        x = sd.fakedir_fromstring(example)
        y = sd.fakedir_fromstring(example_with_comments, parse_comments=False)
        assert x.get_child_names() != y.get_child_names()

class TestFakeDir:
    def test_count_fake_folders(self):
        x = sd.fakedir_fromstring(example)
        assert FDS().count_folders(x.listdir()) == 1

    def test_count_fake_files(self):
        x = sd.fakedir_fromstring(example)
        assert FDS().count_files(x.listdir()) == 3

    def test_sort_fakedir(self):
        x = sd.fakedir_fromstring(example).listdir()
        sort = FDS().sort_dir(x, sort_reverse=True, sort_key=lambda x : x[1])
        sort = [f.name for f in sort]
        correct = ['app.py', 'view.py', 'test', '__init__.py']
        assert sort == correct

    def test_exclude_files_and_reread(self):
        x = sd.fakedir_fromstring(example)
        y = x.seedir(printout=False, exclude_files=r'.*\..*', regex=True)
        z = sd.fakedir_fromstring(y)
        assert set(z.get_child_names()) == set(['test'])

    def test_include_files_and_reread(self):
        x = sd.fakedir_fromstring(example)
        y = x.seedir(printout=False, include_files=['app.py', 'view.py'],
                      regex=False)
        z = sd.fakedir_fromstring(y)
        assert set(z.get_child_names()) == set(['app.py', 'view.py', 'test', ])

    def test_delete_string_names(self):
        x = sd.randomdir()
        x.delete(x.get_child_names())
        assert len(x.listdir()) == 0

    def test_delete_objects(self):
        x = sd.randomdir()
        x.delete(x.listdir())
        assert len(x.listdir()) == 0

    def test_set_parent(self):
        x = sd.fakedir_fromstring(example)
        x['test/test_app.py'].parent = x
        assert 'test_app.py' in x.get_child_names()

    def test_walk_apply(self):
        def add_0(f):
            f.name += ' 0'
        x = sd.fakedir_fromstring(example)
        x.walk_apply(add_0)
        last_chars = [f[-1] for f in x.get_child_names()]
        assert set(last_chars) == set('0')

    def test_depth_setting(self):
        x = sd.fakedir_fromstring(example)
        x['test'].create_folder('A')
        x['test/A'].create_folder('B')
        x['test/A/B'].create_file('boris.txt')
        obj = x['test/A/B/boris.txt']
        depth_a = obj.depth
        obj.parent = x
        depth_b = obj.depth
        assert (depth_a, depth_b) == (4, 1)

    def test_randomdir_seed(self):
        x = sd.randomdir(seed=4.21)
        y = sd.randomdir(seed=4.21)
        assert x.get_child_names() == y.get_child_names()

    def test_empty_fakedir_has_no_children(self):
        x = sd.FakeDir('BORIS')
        assert len(x.get_child_names()) == 0

    def test_populate_adds_children(self):
        x = sd.FakeDir('BORIS')
        sd.populate(x)
        assert len(x.get_child_names()) > 0

    def test_copy_equal(self):
        x = sd.randomdir(seed=7)
        y = x.copy()
        assert x.seedir(printout=False) == y.seedir(printout=False)

    def test_copy_unlinked(self):
        def pallindrome(f):
            f.name = f.name + f.name[::-1]
        x = sd.fakedir_fromstring(large_example)

        before = x.seedir(printout=False)

        y = x.copy()
        y.walk_apply(pallindrome)

        after = x.seedir(printout=False)
        assert before == after

class TestMask:
    def test_mask_no_folders_or_files(self):
        def foo(x):
            if os.path.isdir(x) or os.path.isfile(x):
                return False

        s = sd.seedir(testdir, printout=False, depthlimit=2, itemlimit=10, mask=foo,)
        s = s.split('\n')
        assert len(s) == 1

    def test_mask_always_false(self):
        def bar(x):
            return False
        s = sd.seedir(testdir, printout=False, depthlimit=2, itemlimit=10, mask=bar)
        s = s.split('\n')
        assert len(s) == 1

    def test_mask_fakedir_fromstring(self):
        x = sd.fakedir_fromstring(example)
        s = x.seedir(printout=False, mask=lambda x : not x.name[0] == '_',
                     style='spaces', indent=4)
        assert no_init == s

    def test_mask_fakedir(self):
        def foo(x):
            if os.path.isdir(x) or os.path.isfile(x):
                return False
        f = sd.fakedir(testdir, mask=foo)
        assert len(f.listdir()) == 0

class TestFolderStructure:

    def test_many_randomdirs(self):
        seeds = range(1000)
        results = []
        for i in seeds:
            r = sd.randomdir(seed=i)
            s = r.seedir(printout=False)
            f = sd.fakedir_fromstring(s)
            results.append(s == f.seedir(printout=False))
        assert all(results)

    def test_itemlimit0_nobeyond(self):
        ans = limit0_nobeyond
        f = sd.fakedir_fromstring(large_example)
        s = f.seedir(printout=False, itemlimit=0)
        assert ans == s

    def test_depthlimit0_nobeyond(self):
        ans = limit0_nobeyond
        f = sd.fakedir_fromstring(large_example)
        s = f.seedir(printout=False, depthlimit=0)
        assert ans == s

    def test_itemlimit0_beyond_content(self):
        ans = limit0_beyond_content
        f = sd.fakedir_fromstring(large_example)
        s = f.seedir(printout=False, itemlimit=0, beyond='content')
        assert ans == s

    def test_depthlimit0_beyond_content(self):
        ans = limit0_beyond_content
        f = sd.fakedir_fromstring(large_example)
        s = f.seedir(printout=False, depthlimit=0, beyond='content')
        assert ans == s

    def test_depthlimit1(self):
        ans = depthlimit1
        f = sd.fakedir_fromstring(large_example)
        s = f.seedir(printout=False, depthlimit=1)
        assert ans == s

    def test_depthlimit1_beyond_content(self):
        ans = depthlimit1_beyond_content
        f = sd.fakedir_fromstring(large_example)
        s = f.seedir(printout=False, depthlimit=1, beyond='content')
        assert ans == s

    def test_depthlimit1_beyond_content_exclude(self):
        ans = depthlimit1_beyond_content_exclude
        f = sd.fakedir_fromstring(large_example)
        s = f.seedir(printout=False,
                      depthlimit=1,
                      beyond='content',
                      exclude_files=r'.*\.txt',
                      regex=True)
        assert ans == s

    def test_complex_sort(self):
        ans = complex_sort
        params = dict(sort=True, sort_reverse=True,
                      sort_key = lambda x: len(x), first='files')
        f = sd.fakedir_fromstring(large_example)
        s = f.seedir(printout=False,**params)
        assert ans == s

    def test_complex_inclusion(self):
        ans = complex_inclusion
        params = dict(include_folders=['sandal', 'scrooge', 'pedantic'],
                      exclude_folders='sandal',
                      exclude_files='^Vogel',
                      include_files='^.[oi]',
                      regex=True)
        f = sd.fakedir_fromstring(large_example)
        s = f.seedir(printout=False,**params)
        assert ans == s

class TestFormatter:

    def test_formatter_beyond(self):

        def fmt(p):

            d = {'split':'->', 'final':'->'}

            return d

        ans = fmt_notbeyond
        f = sd.fakedir_fromstring(large_example)
        s = f.seedir(formatter=fmt, itemlimit=1, beyond='content', printout=False)
        assert s == ans


    def test_formatter_no_return(self):

        f = sd.fakedir_fromstring(large_example)
        s = f.seedir(formatter=lambda x: None, printout=False)
        assert s == large_example

    def test_expand_one_folder_sticky(self):

        def fmt(p):

            d = {}
            if p.name == 'scrooge':
                d['depthlimit'] = None

            return d

        ans = fmt_expand_single
        f = sd.fakedir_fromstring(large_example)
        s = f.seedir(formatter=fmt, depthlimit=1, sticky_formatter=True, printout=False)
        assert ans == s

    def test_expand_one_folder_nosticky(self):

        def fmt(p):

            d = {}
            if p.name == 'scrooge':
                d['depthlimit'] = None

            return d

        ans = fmt_expand_single_partial
        f = sd.fakedir_fromstring(large_example)
        s = f.seedir(formatter=fmt, depthlimit=1, printout=False)
        assert ans == s

    def test_mask_with_fmt(self):

        def fmt(p):

            d = {}

            def case1(p):
                return not p.name.endswith('.txt')

            def case2(p):
                return p.name.endswith('.txt')

            if p.depth == 0:
                d['mask'] = case1

            else:
                d['mask'] = case2

            return d

        ans = fmt_with_mask
        f = sd.fakedir_fromstring(large_example)
        s = f.seedir(formatter=fmt, printout=False)
        assert s == ans

class TestTupleItemLimit:

    def count_folder_children(self, f):
        output = []
        foo = lambda x: output.append(len([a for a in f.listdir() if a.isdir()]))
        f.walk_apply(foo)
        return output

    def count_file_children(self, f):
        output = []
        foo = lambda x: output.append(len([a for a in f.listdir() if not a.isdir()]))
        f.walk_apply(foo)
        return output

    def make_letter_example(self):
        f = sd.FakeDir('example')
        f.create_folder(['a', 'b', 'c', 'd'])
        f.create_file(['e', 'f', 'g', 'h'])
        return f

    def test_None_None(self):
        f = sd.fakedir_fromstring(large_example)
        normal = f.seedir(printout=False)
        test = f.seedir(itemlimit=(None, None), printout=False)
        assert normal == test

    def test_None_1(self):
        start = sd.fakedir_fromstring(large_example)
        s = start.seedir(itemlimit=(None, 1), first='files', printout=False)
        end = sd.fakedir_fromstring(s)
        ans = self.count_file_children(end)
        assert all([x <= 1 for x in ans])

    def test_1_None(self):
        start = sd.fakedir_fromstring(large_example)
        s = start.seedir(itemlimit=(1, None), first='folders', printout=False)
        end = sd.fakedir_fromstring(s)
        ans = self.count_folder_children(end)
        assert all([x <= 1 for x in ans])

    def test_1_1(self):
        start = sd.fakedir_fromstring(large_example)
        s = start.seedir(itemlimit=(1, None), first='folders', printout=False)
        end = sd.fakedir_fromstring(s)
        ans = self.count_folder_children(end)
        assert all([x <= 1 for x in ans])

    def test_filter_letter_example_2_2(self):
        e = self.make_letter_example()
        s = e.seedir(itemlimit=(2, 2), sort=True, printout=False)
        f = sd.fakedir_fromstring(s)
        assert set(f.get_child_names()) == set(['a', 'b', 'e', 'f'])

    def test_filter_letter_example_2_None(self):
        e = self.make_letter_example()
        s = e.seedir(itemlimit=(2, None), sort=True, printout=False)
        f = sd.fakedir_fromstring(s)
        assert set(f.get_child_names()) == set(['a', 'b', 'e', 'f', 'g', 'h'])

    def test_filter_letter_example_None_0(self):
        e = self.make_letter_example()
        s = e.seedir(itemlimit=(None, 0), sort=True, printout=False)
        f = sd.fakedir_fromstring(s)
        assert set(f.get_child_names()) == set(['a', 'b', 'c', 'd'])

    def test_filter_letter_example_0_0(self):
        e = self.make_letter_example()
        s = e.seedir(itemlimit=(0, 0), sort=True, printout=False)
        f = sd.fakedir_fromstring(s)
        assert len(f.get_child_names()) == 0

    def test_works_with_list(self):
        f = sd.fakedir_fromstring(large_example)
        s = f.seedir(itemlimit=[0, None], printout=False)
        split = s.split('\n')
        endswithtxt = [x.endswith('txt') for x in split[1:]]
        assert all(endswithtxt)

class TestErrorHandlingArgs:

    def test_handle_errors_correct_type(self):
        x = ErrorRaisingFDS()
        f = sd.fakedir_fromstring(large_example)
        s = x(f, printout=False,
              acceptable_listdir_errors=FakedirError,
              denied_string=' [ACCESS DENIED]')
        assert s == large_example_access_denied

    def test_handle_errors_incorrect_type(self):
        x = ErrorRaisingFDS()
        f = sd.fakedir_fromstring(large_example)
        with pytest.raises(sd.errors.FakedirError):
            _ = x(f, printout=False,
                  acceptable_listdir_errors=PermissionError,
                  denied_string=' [ACCESS DENIED]')

    def test_handle_errors_None(self):
        x = ErrorRaisingFDS()
        f = sd.fakedir_fromstring(large_example)
        with pytest.raises(sd.errors.FakedirError):
            _ = x(f, printout=False,
                  acceptable_listdir_errors=None,
                  denied_string=' [ACCESS DENIED]')

    def test_handle_errors_tuple(self):
        x = ErrorRaisingFDS()
        f = sd.fakedir_fromstring(large_example)
        s = x(f, printout=False,
              acceptable_listdir_errors=(FakedirError, PermissionError),
              denied_string=' [ACCESS DENIED]')
        assert s == large_example_access_denied

    def test_handle_errors_different_string(self):
        x = ErrorRaisingFDS()
        f = sd.fakedir_fromstring(large_example)
        s = x(f, printout=False,
              acceptable_listdir_errors=FakedirError,
              denied_string='<- BUMMER!')
        assert s == large_example_bummer




# -*- coding: utf-8 -*-

# Note: this module fails doctesting

"""
This module provides code for producing folder structure strings for directories.
Currently, the only tool here is `seedir()`, the primary function of the
package `seedir`.  This returns or prints the folder structure for a given path.
The main algorithm for determining the folder structure string is within the
<code>seedir.folderstructure.FolderStructure</code> class.
"""

import os
import pathlib

from seedir.folderstructure import PathlibStructure, RealDirStructure

def _parse_path(path):
    '''Helper function added to parse the input to `seedir.realdir.seedir()`.
    Detects strings (paths) or pathlib objects.'''

    if isinstance(path, str):
        path = os.path.abspath(os.path.expanduser(path))
    elif isinstance(path, pathlib.Path):
        path = path.expanduser().resolve()
    else:
        raise TypeError(f"Can only parse str or pathlib.Path, not {type(path)}.")

    return path

def seedir(path=None, style='lines', printout=True, indent=2, uniform=None,
           anystart=None, anyend=None, depthlimit=None, itemlimit=None,
           beyond=None, first=None, sort=False, sort_reverse=False,
           sort_key=None, include_folders=None, exclude_folders=None,
           include_files=None, exclude_files=None, regex=False, mask=None,
           formatter=None, sticky_formatter=False,
           acceptable_listdir_errors=PermissionError,
           denied_string=' [ACCESS DENIED]', **kwargs):
    '''

    Primary function of the seedir package: generate folder trees for
    computer directories.

    ## EXAMPLES

    ```
    >>> import seedir as sd

    ```

    Make a basic tree diagram:

    ```
    >>> c = 'path/to/doc'
    >>> sd.seedir(c)
    doc/
    ├─_static/
    │ ├─embedded/
    │ │ ├─deep_file
    │ │ └─very/
    │ │   └─deep/
    │ │     └─folder/
    │ │       └─very_deep_file
    │ └─less_deep_file
    ├─about.rst
    ├─conf.py
    └─index.rst

    ```

    Select different styles for the tree:

    ```
    >>> sd.seedir(c, style='dash')
    doc/
    |-_static/
    | |-embedded/
    | | |-deep_file
    | | |-very/
    | |   |-deep/
    | |     |-folder/
    | |       |-very_deep_file
    | |-less_deep_file
    |-about.rst
    |-conf.py
    |-index.rst

    ```

    Sort the folder contents, separting folders and files:

    ```
    >>> sd.seedir(c, sort=True, first='files')
    doc/
    ├─about.rst
    ├─conf.py
    ├─index.rst
    └─_static/
      ├─less_deep_file
      └─embedded/
        ├─deep_file
        └─very/
          └─deep/
            └─folder/
              └─very_deep_file

    ```

    Limit the folder depth or items included:

    ```
    >>> sd.seedir(c, depthlimit=2, itemlimit=1)
    doc/
    ├─_static/
    │ ├─embedded/
    │ └─less_deep_file
    └─about.rst

    ```

    Include or exclude specific items (with or without regular expressions):

    ```
    >>> sd.seedir(c, exclude_folders='_static')
    doc/
    ├─about.rst
    ├─conf.py
    └─index.rst

    ```

    Parameters
    ----------
    path : str, pathlib.Path, or None, optional
        System path of a directory.  If None, current working directory is
        used.  The path can be either a string path or a pathlib object.
        In both cases, the path is converted to an absolute path, and the
        tilde (~) is expanded.
    style : 'lines', 'dash', 'arrow', 'spaces', 'plus', or 'emoji', optional
        Style to use. The default is `'lines'`.  A style determines the set
        of characters ("tokens") used to represent the base structure of
        the directory (e.g. which items belong to which folders, when items
        are the last member of a folder, etc.).  The actual tokens being used
        by each style can be viewed with `seedir.printing.get_styleargs()`.
    printout : bool, optional
        Print the folder structure in the console. The default is `True`.  When
        `False`, the folder diagram is returned as a string.
    indent : int (>= 0), optional
        Number of spaces separating items from their parent folder.
        The default is `2`.
    uniform : str or None, optional
        Characters to use for all tokens when creating the tree diagram.
        The default is `None`.  When not `None`, the extend, space, split, and
        final tokens are replaced with `uniform` (the `'spaces'` style is
        essentially `uniform = '  '`).
    anystart : str or None, optional
        Characters to prepend before any item (i.e. folder or file).  The
        default is `None`.  Specific starts for folders and files can be
        specified (see `**kwargs`).
    anyend : str or None, optional
        Characters to append after any item (i.e. folder or file).  The
        default is `None`.  Specific ends for folders and files can be
        specified (see `**kwargs`).
    depthlimit : int or None, optional
        Limit the depth of folders to traverse.  Folders at the `depthlimit` are
        included, but their contents are not shown (with the exception of the
        beyond parameter being specified).  The default is `None`, which can
        cause exceptionally long runtimes for deep or extensive directories.
    itemlimit : int or None, optional
        Limit the number of items in a directory to show.  Items beyond the
        `itemlimit` can be expressed using the `beyond` parameter.  The files and
        folders left out are determined by the sorting parameters
        (`sort`, `sort_reverse`, `sort_key`).  The default is `None`.
    beyond : str ('ellipsis', 'content' or a string starting with an underscore) or None, optional
        String to indicate directory contents beyond the `itemlimit` or the
        `depthlimit`.  The default is `None`.  Options are: `'ellipsis'` (`'...'`),
        `'content'` or `'contents'` (the number of files and folders beyond), or
        a string starting with `'_'` (everything after the leading underscore
        will be returned)
    first : 'files', 'folders', or None, optional
        Sort the directory so that either files or folders appear first.
        The default is `None`.
    sort : bool, optional
        Sort the directory. With no other specifications, the sort will be a
        simple alphabetical sort of the item names, but this can be altered
        with the `first`, `sort_reverse`, and `sort_key parameters`.
        The default is `False`.
    sort_reverse : bool, optional
        Reverse the sorting determined by `sort` or `sort_key`.
        The default is `False`.
    sort_key : function, optional
        Key to use for sorting file or folder names, akin to the `key` parameter
        of the builtin `sorted()` or `list.sort()`. The function should take a
        string as an argument. The default is `None`.
    include_folders, exclude_folders, include_files, exclude_files : str, list-like, or None, optional
        Folder / file names to include or exclude. The default is `None`.  By
        default, these are interpreted literally.  Pass `regex=True` for
        using regular expressions.
    regex : bool, optional
        Interpret the strings of include/exclude file/folder arguments as
        regular expressions. The default is `False`.
    mask : function, optional
        Function for filtering items.  Absolute paths of each individual item
        are passed to the `mask` function.  If `True` is returned, the
        item is kept.  The default is `None`.  The type of the object
        passed to `mask` corresponds with that passed as input:
        `str` or `pathlib.Path`.
    formatter : function, optional
        Function for customizing the directory printing logic and style
        based on specific folders & files.  When passed, the formatter
        is called on each item in the file tree, and the current arguments
        are updated based what is returned.

        The formatter function should accept a system path as a
        single argument (either relative or absolute, depending on what is passed
        to the `path` argument), and it should return either a dictionary or None.
        The dictionary should have names of arguments as keys and their respective
        setting as values.

        The following options can meaningfully be toggled by passing a formatter
        function: `depthlimit`, `itemlimit`, `beyond`, `first`, `sort`, `sort_reverse`,
        `sort_key`, `include_folders`, `regex`, `mask`, as well as any seedir token
        keywords (`extend`, `space`, `split`, `final`, `folderstart`, `filestart`,
        `folderend`, `fileend`).

        Note that in version 0.3.0, formatter could only be used to update
        the style tokens.  It can now be used to udpate those as well as the other
        arguments listed above.

        If None is returned by formatter, the tokens will be set by `style`.

        Note that items exlcuded by the inclusion/exclusion arguments (or the
        `mask`) *will not* be seen by formatter.  Similarly, any folder tree
        entries created by the `beyond` argument *will not* be seen by formatter.

        The type of the object passed to `formatter` corresponds with that
        passed as input: `str` or `pathlib.Path`.

    sticky_formatter : bool, optional
        When True, updates to argumnts made by the `formatter` (see above)
        will be permanent.  Thus, if arguments are updated when the `formatter`
        is called on a folder, its children will (recursively) inherit
        those new arguments.
    acceptable_listdir_errors : Exception, tuple, or None

        **New in v0.5.0**

        Set errors which, when raised when listing the contents of a folder,
        do not halt traversal.  This parameter was added to allow
        folders to be skipped when a `PermissionError` is raised.  For folders
        which raise an acceptable error, an additional string is added to their
        entry in the diagram (see `denied_string`).

        Providing one Exception causes only that type of Exception to be ignored.
        Multiple Exception types can be handled by passing mutple Exceptions
        as a tuple.  Pass `None` to not allow any Exceptions (in this case,
        an error is raised).

        Exception types which are not provided will still be raised.

        The default is `PermissionError`, which will skip folders
        for which the caller does not have permissions to access.

    denied_string : str

        String tag to signify that a folder was not able to be traversed
        due to one of the `acceptable_listdir_errors` being raised.  This
        is a string added after the folder name (and `folderend`) strings.
        The default is `" [ACCESS DENIED]"`.

    **kwargs : str
        Specific tokens to use for creating the file tree diagram.  The tokens
        use by each builtin style can be seen with `seedir.printing.get_styleargs()`.
        Valid options are `extend` (characters to show the extension of a directory
        while its children are traversed), `space` (character to provide the
        correct indentation of an item when some of its parent / grandparent
        directories are completely traversed), `split` (characters to show a
        folder or file within a directory, with more items following),
        `final` (characters to show a folder or file within a directory,
        with no more items following), `folderstart` (characters to prepend
        before any folder), `filestart` (characters to preppend before any
        file), `folderend` (characters to append after any folder), and
        `fileend` (characters to append after any file). The following shows
        the default tokens for the `'lines'` style:

            >>> import seedir as sd
            >>> sd.get_styleargs('lines')
            {'split': '├─', 'extend': '│ ', 'space': '  ', 'final': '└─', 'folderstart': '', 'filestart': '', 'folderend': '/', 'fileend': ''}

        All default style tokens are 2 character strings, except for
        the file/folder start/end tokens.  Style tokens from `**kwargs` are not
        affected by the indent parameter.  The `uniform`, `anystart`, and
        `anyend` parameters can be used to affect multiple style tokens.

    Notes
    -------

    The parameter `slash` was deprecated in 0.5.0.  Pass `folderend` as
    an additional keyword argument instead.

    Returns
    -------
    s (str) or None
        The tree diagram (as a string) or `None` if `prinout = True`, in which
        case the tree diagram is printed in the console.

    '''

    if path is None:
        path = os.getcwd()

    path = _parse_path(path)

    # call
    args = dict(style=style,
                printout=printout,
                indent=indent,
                uniform=uniform,
                anystart=anystart,
                anyend=anyend,
                depthlimit=depthlimit,
                itemlimit=itemlimit,
                beyond=beyond,
                first=first,
                sort=sort,
                sort_reverse=sort_reverse,
                sort_key=sort_key,
                include_folders=include_folders,
                exclude_folders=exclude_folders,
                include_files=include_files,
                exclude_files=exclude_files,
                regex=regex,
                mask=mask,
                formatter=formatter,
                sticky_formatter=sticky_formatter,
                acceptable_listdir_errors=acceptable_listdir_errors,
                denied_string=denied_string,
                **kwargs)

    structure = PathlibStructure() if isinstance(path, pathlib.Path) else RealDirStructure()

    return structure(path, **args)



# -*- coding: utf-8 -*-
"""
Unit tests for seedir.

Previously written for unittests, and updated for pytest.

Test methods MUST start with "test"
"""

import os

import seedir as sd

# realdir for testing on
testdir = os.path.dirname(os.path.abspath(__file__))

# ---- Test cases

print('\n--------------------'
      '\n\nTesting seedir.seedir() against {}\n\n'
      '--------------------'
      '\n'.format(testdir))

print('Basic seedir (depthlimit=2, itemlimit=10):\n')
sd.seedir(testdir, depthlimit=2, itemlimit=10)

print('\nDifferent Styles (depthlimit=1, itemlimit=5):')
for style in sd.STYLE_DICT.keys():
    print('\n{}:\n'.format(style))
    sd.seedir(testdir, style=style, depthlimit=1, itemlimit=5)

print('\nCustom Styles (depthlimit=1, itemlimit=5):')
sd.seedir(testdir, depthlimit=1, itemlimit=5, space='>>',
          split='>>', extend='II', final='->',
          folderstart='Folder: ', filestart='File: ')

print('\nDifferent Indents (depthlimit=1, itemlimit=5):')
for i in list(range(3)) + [8]:
    print('\nindent={}:\n'.format(str(i)))
    sd.seedir(testdir, depthlimit=1, itemlimit=5, indent=i)

print('\nItems Beyond Limit (depthlimit=1, itemlimit=1, beyond="content")')
sd.seedir(testdir, itemlimit=1, beyond='content')





#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import argparse
import os

from .realdir import seedir

helptxt = """
Help for the seedir CLI.  Function for printing a folder tree
structure, similar to UNIX `tree`.  Directs all arguments to `seedir.seedir()`.

Note that this has been revamped for version v0.3.1.  Specifically, parsing
was moved from getopts to argparse.  Options should now all explicitly
reference parameters of `seedir.seedir()`.

Not all seedir arguments are accepted, namely ones which expect Python
callables and ones which alter the style.  The latter may be added
in a future version.

Options (short/long):

    -b / --beyond BEYOND
        Way to represent items past the depthlimit or itemlimit.
        Options are 'content', 'ellipsis', or explicit string starting
        with an underscore.  Not set by default.

    -d / --depthlimit DEPTHLIMIT
        Integer limit on depth of folders to enter. Default: None

    -f / --first FIRST
        Sort the directory so that either files or folders appear first.
        Options are "folders" or "files".

    -h / --help
        Show this message and exit.

    -i / --itemlimit ITEMLIMIT
        Integer limit on the number of items to include per directory.
        Default: None

    -p / --path PATH
        Dir to see.  Default: current directory (os.getcwd())

    -r / --sort_reverse
        Reverse the sort.

    -s / --sort
        Sort the contents of each directory by name.

    -t / --indent INDENT
        Amount to indent.  Default: 2

    -y / --style STYLE
        seedir style to use.  Default: "lines"

Options (long only):

    --include_folders, --include_files, --exclude_folders, --exclude_files ...
        Folder / file names to include or exclude.  Not set by default. Pass
        multiple items as space-separated values.

    --regex
        Interpret include/exclude folder/file arguments as Python regex.
"""

def parse():

    """Parse command line arguments with argparse."""

    parser = argparse.ArgumentParser(add_help=False)

    parser.add_argument('-b', '--beyond', default=None)
    parser.add_argument('-d', '--depthlimit', default=None, type=int)
    parser.add_argument('-f', '--first', default=None)
    parser.add_argument('-h', '--help', action='store_true')
    parser.add_argument('-i', '--itemlimit', default=None, type=int)
    parser.add_argument('-p', '--path', default=os.getcwd())
    parser.add_argument('-r', '--sort_reverse', action='store_true')
    parser.add_argument('-s', '--sort', action='store_true')
    parser.add_argument('-t', '--indent', default=2, type=int)
    parser.add_argument('-y', '--style', default='lines')

    parser.add_argument('--include_folders', default=None, nargs='+')
    parser.add_argument('--exclude_folders', default=None, nargs='+')
    parser.add_argument('--include_files', default=None, nargs='+')
    parser.add_argument('--exclude_files', default=None, nargs='+')
    parser.add_argument('--regex', action='store_true')

    return parser.parse_args()

def main():

    """Parses arguments and passes them to `seedir.seedir()`"""

    args = parse()
    if args.help:
        print(helptxt)
        return

    kwargs = vars(args)
    del kwargs['help']
    seedir(**kwargs)

if __name__ == '__main__':
    main()




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Setup for seedir.

@author: Tom Earnest
"""

from os import path

from setuptools import setup

# read the contents of your README file
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(name='seedir',
      version='0.5.0',
      description='Package for creating, editing, and reading folder tree diagrams.',
      url='https://github.com/earnestt1234/seedir',
      author='Tom Earnest',
      author_email='earnestt1234@gmail.com',
      license='MIT',
      packages=['seedir'],
      install_requires=['natsort'],
      extras_require={'emoji': ['emoji']},
      include_package_data=True,
      zip_safe=False,
      long_description=long_description,
      long_description_content_type='text/markdown',
      entry_points = {
        'console_scripts': ['seedir=seedir.__main__:main'],
        })



# -*- coding: utf-8 -*-
"""
General module of resources and helpers for printing and making folder trees
in seedir.

"""

__pdoc__ = {'is_match': False,
            'format_indent': False,
            'words': False}



import copy
import os
import re


STYLE_DICT = {
    'lines': {'split':'├─',
              'extend':'│ ',
              'space':'  ',
              'final':'└─',
              'folderstart':'',
              'filestart':'',
              'folderend': '/',
              'fileend': ''},
    'dash':  {'split':'|-',
              'extend':'| ',
              'space':'  ',
              'final':'|-',
              'folderstart':'',
              'filestart':'',
              'folderend': '/',
              'fileend': ''},
    'spaces':{'split':'  ',
              'extend':'  ',
              'space':'  ',
              'final':'  ',
              'folderstart':'',
              'filestart':'',
              'folderend': '/',
              'fileend': ''},
    'plus':  {'split':'+-',
              'extend':'| ',
              'space':'  ',
              'final':'+-',
              'folderstart':'',
              'filestart':'',
              'folderend': '/',
              'fileend': ''},
    'arrow': {'split':'  ',
              'extend':'  ',
              'space':'  ',
              'final':'  ',
              'folderstart':'>',
              'filestart':'>',
              'folderend': '/',
              'fileend': ''}
    }
'''"Tokens" used to create folder trees in different styles'''

try:
    import emoji
    STYLE_DICT["emoji"] = {
        'split':'├─',
        'extend':'│ ',
        'space':'  ',
        'final':'└─',
        'folderstart':emoji.emojize(':file_folder:' + ' '),
        'filestart':emoji.emojize(':page_facing_up:' + ' '),
        'folderend': '/',
        'fileend': ''
    }
except ImportError:
    pass


filepath = os.path.dirname(os.path.abspath(__file__))
wordpath = os.path.join(filepath, 'words.txt')
with open(wordpath, 'r') as wordfile:
    words = [line.strip() for line in wordfile.readlines()]
"""List of dictionary words for seedir.fakedir.randomdir()"""

# functions

def is_match(pattern, string, regex=True):
    '''Function for matching strings using either regular expression
    or literal interpretation.'''
    if regex:
        return bool(re.search(pattern, string))
    else:
        return pattern == string

def get_styleargs(style):
    '''
    Return the string tokens associated with different styles for printing
    folder trees with `seedir.realdir.seedir()`.

    Parameters
    ----------
    style : str
        Style name.  Current options are `'lines'`, `'spaces'`, `'arrow'`,
        `'plus'`, `'dash'`, or `'emoji'`.

    Raises
    ------
    ValueError
        Style not recognized.

    ImportError
        'emoji' style requested but emoji package not installed.

    Returns
    -------
    dict
        Dictionary of tokens for the given style.

    '''
    if style not in STYLE_DICT and style == 'emoji':
        error_text = 'style "emoji" requires "emoji" to be installed'
        error_text += ' (pip install emoji) '
        raise ImportError(error_text)
    elif style not in STYLE_DICT:
        error_text = 'style "{}" not recognized, must be '.format(style)
        error_text += 'lines, spaces, arrow, plus, dash, or emoji'
        raise ValueError(error_text)
    else:
        return copy.deepcopy(STYLE_DICT[style])

def format_indent(style_dict, indent=2):
    '''
    Format the indent of style tokens, from seedir.STYLE_DICT or returned
    by seedir.get_styleargs().

    Note that as of v0.3.1, the dictionary is modified in place,
    rather than a new copy being created.

    Parameters
    ----------
    style_dict : dict
        Dictionary of style tokens.
    indent : int, optional
        Number of spaces to indent. The default is 2.  With 0, all tokens
        become the null string.  With 1, all tokens are only the first
        character.  With 2, the style tokens are returned unedited.  When >2,
        the final character of each token (excep the file/folder start/end tokens)
        are extened n - indent times, to give a string whose
        length is equal to indent.

    Returns
    -------
    output : dict
        New dictionary of edited tokens.

    '''
    indentable = ['split', 'extend', 'space', 'final']
    if indent < 0 or not isinstance(indent, int):
        raise ValueError('indent must be a non-negative integer')
    elif indent == 0:
        for key in indentable:
            style_dict[key] = ''
    elif indent == 1:
        for key in indentable:
            style_dict[key] = style_dict[key][0]
    elif indent > 2:
        extension = indent - 2
        for key in indentable:
            val = style_dict[key]
            style_dict[key] = val + val[-1] * extension

    return None
